using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace PorygonSharp.DiagnosticHandling
{
    [StructLayout(LayoutKind.Sequential)]
    internal struct InternalDiagnostics
    {
        private readonly byte _hasErrors;

        public bool HasErrors()
        {
            return _hasErrors == 1;
        }
    }

    public class Diagnostics : IDisposable
    {
        private readonly IntPtr _handle;
        private InternalDiagnostics _internal;

        public Diagnostics(IntPtr handle)
        {
            _handle = handle;
            _internal = Marshal.PtrToStructure<InternalDiagnostics>(handle);
        }

        public void Dispose()
        {
            Marshal.FreeHGlobal(_handle);
        }

        public bool HasErrors()
        {
            return _internal.HasErrors();
        }

        public int DiagnosticsCount()
        {
            return GetDiagnosticsCount(_handle);
        }

        public IEnumerable<DiagnosticItem> GetDiagnostics()
        {
            var count = DiagnosticsCount();
            for (var i = 0; i < count; i++)
            {
                var ptr = GetDiagnosticAt(_handle, i);
                yield return new DiagnosticItem(ptr);
            }
        }

        public string GetFullDiagnosticMessage(DiagnosticItem item)
        {
            return GetFullDiagnostic(_handle, item.GetHandle());
        }

        [DllImport("PorygonLang", EntryPoint = "GetDiagnosticsCount", CallingConvention = CallingConvention.Cdecl)]
        private static extern int GetDiagnosticsCount(IntPtr ptr);

        [DllImport("PorygonLang", EntryPoint = "GetDiagnosticAt", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr GetDiagnosticAt(IntPtr ptr, int position);

        [DllImport("PorygonLang", EntryPoint = "GetFullDiagnostic", CallingConvention = CallingConvention.Cdecl)]
        private static extern string GetFullDiagnostic(IntPtr diagnosticsHolder, IntPtr diagnostic);
    }
}