using System;
using System.Collections;
using System.Runtime.InteropServices;

namespace PorygonSharp.EvalValues
{
    public class DictionaryEvalValue
    {
        private readonly IDictionary _dictionary;
        private readonly Type _keyType;
        private readonly Type _valueType;

        private delegate IntPtr GetterDelegate(IntPtr parent, IntPtr index);
        private delegate void SetterDelegate(IntPtr parent, IntPtr index, IntPtr value);
        private delegate IntPtr GetIteratorDelegate(IntPtr parent);
        private delegate int GetLengthDelegate(IntPtr parent);

        // ReSharper disable PrivateFieldCanBeConvertedToLocalVariable
        private readonly GetterDelegate _getter; 
        private readonly SetterDelegate _setter; 
        private readonly GetIteratorDelegate _getIterator; 
        private readonly GetLengthDelegate _getLength; 
        // ReSharper restore PrivateFieldCanBeConvertedToLocalVariable

        public readonly IntPtr GetterPtr;
        public readonly IntPtr SetterPtr;
        public readonly IntPtr GetIteratorPtr;
        public readonly IntPtr GetLengthPtr;

        public DictionaryEvalValue(IDictionary dictionary, Type t)
        {
            _dictionary = dictionary;
            var genericTypes = t.GetGenericArguments();
            _keyType = genericTypes[0];
            _valueType = genericTypes[1];

            _getter = Getter;
            _setter = Setter;
            _getIterator = GetIterator;
            _getLength = GetLength;

            GetterPtr = Marshal.GetFunctionPointerForDelegate(_getter);
            SetterPtr = Marshal.GetFunctionPointerForDelegate(_setter);
            GetIteratorPtr = Marshal.GetFunctionPointerForDelegate(_getIterator);
            GetLengthPtr = Marshal.GetFunctionPointerForDelegate(_getLength);
        }
        
        private IntPtr Getter(IntPtr parent, IntPtr index)
        {
            var val = new EvalValue(index).GetObjectValue();
            return EvalValueCreator.CreateValue(_dictionary[val]).GetPointer();
        }
        
        private void Setter(IntPtr parent, IntPtr index, IntPtr value)
        {
            var key = new EvalValue(index).GetObjectValue();
            var val = new EvalValue(value).GetObjectValue();
            _dictionary[key] = Convert.ChangeType(val, _valueType);
        }
        
        private IntPtr GetIterator(IntPtr parent)
        {
            throw new NotImplementedException();
        }

        private int GetLength(IntPtr parent)
        {
            return _dictionary.Count;
        }

    }
}