using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Runtime.InteropServices;
using PorygonSharp.Utilities;

namespace PorygonSharp.EvalValues
{
    internal class ListEvalValue
    {
        private static readonly ConcurrentDictionary<IntPtr, WeakReference<ListEvalValue>> _cache 
            = new ConcurrentDictionary<IntPtr, WeakReference<ListEvalValue>>();
        
        public IList List { get; }
        private readonly Type _valueType;
        
        private delegate IntPtr GetterDelegate(IntPtr parent, IntPtr index);
        private delegate void SetterDelegate(IntPtr parent, IntPtr index, IntPtr value);
        private delegate IntPtr GetIteratorDelegate(IntPtr parent);
        private delegate int GetLengthDelegate(IntPtr parent);

        // ReSharper disable PrivateFieldCanBeConvertedToLocalVariable
        private readonly GetterDelegate _getter; 
        private readonly SetterDelegate _setter; 
        private readonly GetIteratorDelegate _getIterator; 
        private readonly GetLengthDelegate _getLength; 
        // ReSharper restore PrivateFieldCanBeConvertedToLocalVariable

        public readonly IntPtr ParentCollectionPtr;
        public readonly IntPtr GetterPtr;
        public readonly IntPtr SetterPtr;
        public readonly IntPtr GetIteratorPtr;
        public readonly IntPtr GetLengthPtr;
        
        
        private ListEvalValue(IList list, Type type)
        {
            List = list;
            _valueType = type.IsArray ? type.GetElementType() : type.GetGenericArguments()[0];
            if (_valueType == null)
                throw new ArgumentNullException();

            ParentCollectionPtr = IntPtr.Zero;
            _getter = Getter;
            _setter = Setter;
            _getIterator = GetIterator;
            _getLength = GetLength;

            GetterPtr = Marshal.GetFunctionPointerForDelegate(_getter);
            GetterPtr = Marshal.GetFunctionPointerForDelegate(_getter);
            SetterPtr = Marshal.GetFunctionPointerForDelegate(_setter);
            GetIteratorPtr = Marshal.GetFunctionPointerForDelegate(_getIterator);
            GetLengthPtr = Marshal.GetFunctionPointerForDelegate(_getLength);
            
        }

        public static IntPtr Create(IList list, Type type)
        {
            var v = new ListEvalValue(list, type);
            var ptr = EvalValueCreator.CreateCollectionValue(IntPtr.Zero, v.ParentCollectionPtr, v.GetterPtr,
                v.SetterPtr, v.GetIteratorPtr, v.GetLengthPtr);
            _cache.TryAdd(ptr, new WeakReference<ListEvalValue>(v));
            return ptr;
        }

        internal static ListEvalValue Fetch(IntPtr ptr)
        {
            if (_cache.TryGetValue(ptr, out var r))
            {
                if (r.TryGetTarget(out var v))
                {
                    return v;
                }
            }
            throw new Exception("Can't get list eval value.");
        }

        private IntPtr Getter(IntPtr parent, IntPtr index)
        {
            var val = new EvalValue(index).EvaluateInteger();
            return EvalValueCreator.CreateValue(List[(int) val - 1]).GetPointer();
        }
        
        private void Setter(IntPtr parent, IntPtr index, IntPtr value)
        {
            var key = new EvalValue(index).EvaluateInteger();
            var val = new EvalValue(value);
            List[(int) key - 1] = Caster.Cast(val, _valueType);
        }
        
        private IntPtr GetIterator(IntPtr parent)
        {
            return ListIterator.CreateListIterator(List);
        }

        private int GetLength(IntPtr parent)
        {
            return List.Count;
        }
    }
}