using System;
using System.Collections.Concurrent;
using System.Runtime.InteropServices;

namespace PorygonSharp.EvalValues
{
    internal static class ObjectEvalValueHandler
    {
        private struct PtrObject
        {
            public readonly IntPtr Pointer;
            public readonly WeakReference<object> WeakReference;

            public PtrObject(IntPtr pointer, WeakReference<object> weakReference)
            {
                Pointer = pointer;
                WeakReference = weakReference;
            }
        }
        
        private static readonly ConcurrentDictionary<int, PtrObject> CreatedPointers = new ConcurrentDictionary<int, PtrObject>();

        public static IntPtr GetObjectPtr(object o)
        {
            var hash = o.GetHashCode();
            if (CreatedPointers.TryGetValue(hash, out var ptrObject))
            {
                if (ptrObject.WeakReference.TryGetTarget(out _))
                {
                    return ptrObject.Pointer;
                }
                else
                {
                    CreatedPointers.TryRemove(hash, out _);
                }
            }
            var handle   = GCHandle.Alloc(o, GCHandleType.WeakTrackResurrection);
            var ptr = GCHandle.ToIntPtr(handle);
            ptrObject = new PtrObject(ptr, new WeakReference<object>(o));
            CreatedPointers.TryAdd(o.GetHashCode(), ptrObject);
            return ptrObject.Pointer;
        }
    }
}