using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace PorygonSharp.EvalValues
{
    internal class TableEvalValue
    {
        private readonly IntPtr _handle;
        
        public TableEvalValue(IntPtr ptr)
        {
            _handle = ptr;
        }

        public enum TableType
        {
            Unknown,
            Numerical,
            StringKeyed,
            Dictionary
        }

        public TableType GetTableType()
        {
            return GetTableType(_handle);
        }

        public int GetLength()
        {
            return GetTableLength(_handle);
        }

        public EvalValue IndexTable(uint hash)
        {
            var ptr = IndexTableHash(_handle, hash);
            return new EvalValue(ptr);
        }
        
        public EvalValue IndexTable(EvalValue key)
        {
            var ptr = IndexTableObj(_handle, key.GetPointer());
            return new EvalValue(ptr);
        }

        public IEnumerator<object> EvaluateNumericalEnumerator()
        {
            if (GetTableType() != TableType.Numerical)
                throw new Exception("Can't evaluate table as array. It's not a numerical table");

            var length = GetLength();
            for (uint i = 1; i <= length; i++)
            {
                var val = IndexTable(i);
                yield return val.GetObjectValue();
            }
        }

        public object[] EvaluateNumericalTable()
        {
            var length = GetLength();
            var array = new object[length];
            var c = 0;
            using (var enumerator = EvaluateNumericalEnumerator())
            {
                while (enumerator.MoveNext())
                {
                    array[c] = enumerator.Current;
                    c++;
                }
            }
            return array;
        }


        [DllImport("PorygonLang", EntryPoint = "GetTableType", CallingConvention = CallingConvention.Cdecl)]
        private static extern TableType GetTableType(IntPtr ptr);
        
        [DllImport("PorygonLang", EntryPoint = "GetTableLength", CallingConvention = CallingConvention.Cdecl)]
        private static extern int GetTableLength(IntPtr ptr);

        [DllImport("PorygonLang", EntryPoint = "IndexTableHash", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr IndexTableHash(IntPtr ptr, uint hash);
        
        [DllImport("PorygonLang", EntryPoint = "IndexTableObj", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr IndexTableObj(IntPtr ptr, IntPtr key);

    }
}