using System;
using System.Runtime.InteropServices;
using PorygonSharp.UserData;

namespace PorygonSharp.ScriptType
{
    public class ScriptType
    {
        private readonly IntPtr _ptr;

        public ScriptType(IntPtr ptr)
        {
            _ptr = ptr;
        }

        public TypeClass GetTypeClass()
        {
            return GetTypeClass(_ptr);
        }

        public uint GetUserDataType()
        {
            return GetScriptTypeUserData(_ptr);
        }

        internal Type ResolveType()
        {
            var c = GetTypeClass();
            switch (c)
            {
                case TypeClass.Number: return typeof(long);
                case TypeClass.Bool: return typeof(bool);
                case TypeClass.String: return typeof(string);
                case TypeClass.UserData:
                    var id = GetUserDataType();
                    var t = UserDataHandler.ReverseLookup[id];
                    return t.Type;
                default:
                    throw new ArgumentOutOfRangeException();
            }
        }
        
        
        
        [DllImport("PorygonLang", EntryPoint = "GetTypeClass", CallingConvention = CallingConvention.Cdecl)]
        private static extern TypeClass GetTypeClass(IntPtr t);
        
        [DllImport("PorygonLang", EntryPoint = "GetScriptTypeUserData", CallingConvention = CallingConvention.Cdecl)]
        private static extern uint GetScriptTypeUserData(IntPtr t);

    }
}