using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using PorygonSharp.EvalValues;
using PorygonSharp.UserData;
using PorygonSharp.Utilities;

namespace PorygonSharp
{
    public static class StaticScope
    {
        private static List<object> _variables = new List<object>();
        private static List<EvalValue> _evalValues = new List<EvalValue>();
        public static void RegisterStaticVariable(string name, object o)
        {
            var type = o.GetType();
            var scriptType = ScriptType.ScriptTypeHandler.GetScriptType(type);
            if (!scriptType.HasValue)
                return;
            var hash = name.ScriptHash();
            var value = EvalValueCreator.CreateValue(o);
            _variables.Add(o);
            _evalValues.Add(value);
            RegisterStaticVariable(hash, scriptType.Value, value.GetPointer());
        }

        public static void RegisterStaticEnum(string name, Type type)
        {
            UserDataHandler.RegisterEnumType(name, type);
            var typeHash = UserDataHandler.GetTypeId(type);
            var scriptType = ScriptType.ScriptTypeHandler.GetScriptType(type);
            if (!scriptType.HasValue)
                return;
            var ptr = ObjectEvalValueHandler.GetObjectPtr(Activator.CreateInstance(type));
            var val = new EvalValue(EvalValueCreator.CreateUserDataEvalValue(typeHash, ptr));
            StaticScope.RegisterStaticVariable(name, scriptType.Value, val);
        }
        
        public static void RegisterStaticVariable(string name, IntPtr type, EvalValue val)
        {
            var hash = name.ScriptHash();
            RegisterStaticVariable(hash, type, val.GetPointer());
        }

        
        [DllImport("PorygonLang", EntryPoint = "RegisterStaticVariable", CallingConvention = CallingConvention.Cdecl)]
        private static extern void RegisterStaticVariable(uint hashId, IntPtr scriptType, IntPtr value);
    }
}