using System;
using System.Collections;
using System.Collections.Generic;
using PorygonSharp.EvalValues;

namespace PorygonSharp.Utilities
{
    public static class Caster
    {
        public static object Cast(EvalValue val, Type t)
        {
            var eval = val.GetObjectValue();
            return Cast(eval, t);
        }

        private static object Cast(object o, Type t)
        {
            if (t.IsEnum)
            {
                var underlying = t.GetEnumUnderlyingType();
                return Convert.ChangeType(o, underlying);
            }
            if (o is IConvertible)
            {
                return Convert.ChangeType(o, t);
            }
            if (typeof(IList).IsAssignableFrom(t) && o is object[] arr)
            {
                return CastList(arr, t);
            }
            return o;
        }

        private static object CastList(IReadOnlyList<object> o, Type t)
        {
            IList arr;
            Type elementType;
            if (t.IsArray)
            {
                elementType = t.GetElementType();
                arr = Array.CreateInstance(elementType, o.Count);
            }
            else
            {
                elementType = t.GetGenericArguments()[0];
                arr = (IList)Activator.CreateInstance(t, o.Count);
            }
            for (var i = 0; i < o.Count; i++)
            {
                arr[i] = Cast(o[i], elementType);
            }

            return arr;
        }
    }
}